// @flow
import _ from 'lodash/fp';
import yargs from 'yargs';
import debug from 'debug';

import type { Task } from './tasks';

import api from './api';
import tasks from './tasks';

const d = debug('cli');

type Args = {
  argv: Object,
};

// Those options from the argv object are not very interesting to us.
const argvBoilerplate = ['_', 'version', 'v', 'help', 'h', '$0', 'opts'];

// Turn a name into a standard Environment variable name, e.g.:
// cool-bananas => CAC_COOL_BANANAS
const toEnvName = _.flow(_.snakeCase, _.toUpper, (n) => `CAC_${n}`);

// Place a variable into the process environment if it doesn't exist yet.
const setEnvVar = (name: string, value: any): void => {
  const envName = toEnvName(name);
  if (!_.has(envName, process.env)) { process.env[envName] = value; }
};

// Populate the command line parser with task specific commands and options.
const taskArgs = (commands: Array<Task>, args: Args): Args =>
  _.reduce((memo, command) => {
    const cmdModule = _.pick(['builder', 'handler'], command);
    return memo.command(command.cmd, command.description, cmdModule);
  }, args, commands);

// Transition from processing the args object to processing the arguments.
const toArgv = (args: Args): Object => args.argv;

// Copy the arguments into the environment.
const setEnv = (argv: Object): Object => {
  _(argv).omit(argvBoilerplate)
         // Only choose options that have a full name and have a value.
         .pickBy((v: any, k: string) => v && _.size(k) !== 1)
         .forEach((v: any, k: string) => setEnvVar(k, v));
  return argv;
};

// Collect all command arguments and handle them.
const handleArgs = (commands: Array<Task>, args: Args): Object =>
  _.flow([_.curry(taskArgs)(commands),
          toArgv,
          setEnv])(args);

export const cli = (argv: Array<string>): Args => _(yargs(argv)
  .option('c', {
    alias: 'config',
    nargs: 1,
    describe: 'Specify a configFile',
    type: 'string',
    default: './.cac.json',
  })
  .usage('Usage: $0 -c <path> <command> [<args>]')
  .alias('v', 'version')
  .version(api.version)
  .describe('v', 'Show version info.')
  .help('h')
  .alias('h', 'help')
  .showHelpOnFail(false, 'Specify -h or --help for available options.'))
  // From here on starts the lodash chain pipeline. Note the 2 parentheses the
  // line before!!
  .thru(_.curry(handleArgs)(tasks))
  .tap(d)
  .value();
