import _ from 'lodash/fp';
import Immutable from 'immutable';
import document from '../lib/document';

const docJs = (data) => _.merge(document.defaults, data);

describe('Document type', function () {
  it('can deeply convert to JavaScript', function () {
    const m = document.document({a: {b: 1}, c: 2});

    document.toJS(m).should.eql(docJs({a: {b: 1}, c: 2}));
  });

  it('can deeply convert from JavaScript', function () {
    const m = document.document({a: 1});

    document.toJS(m).should.eql(docJs({a: 1}));
  });

  it('has an identity element', function () {
    const id1 = document.id();
    const id2 = document.id();

    Immutable.is(id1, id2).should.equal(true);
  });

  it('aliases mempty to the identity element', function () {
    const id1 = document.id();
    const id2 = document.mempty();

    Immutable.is(id1, id2).should.equal(true);
  });

  it('maps a function over the document', function () {
    const m1 = document.document({a: 1, b: 2});
    const inc = (v) => (_.isInteger(v) ? v + 1 : v);
    const mapped = document.fmap(inc, m1);

    document.toJS(mapped).should.eql(docJs({a: 2, b: 3}));
  });

  it('has a binary append operation', function () {
    const m1 = document.document({a: 1});
    const m2 = document.document({b: 2});
    const m3 = document.mappend(m1, m2);

    document.toJS(m3).should.eql(docJs({a: 1, b: 2}));
  });

  it('folds a Immutable.List<Document>', function () {
    const ms = Immutable.List([document.document({a: 1}),
                               document.document({b: 2})]);
    const m = document.mconcat(ms);

    document.toJS(m).should.eql(docJs({a: 1, b: 2}));
  });


  it('folds an Array<Document>', function () {
    const ms = [document.document({a: 1}), document.document({b: 2})];
    const m = document.mconcat(ms);

    document.toJS(m).should.eql(docJs({a: 1, b: 2}));
  });

  it('holds for the first functor law', function () {
    // fmap identity = identity
    const m = document.document();
    const firstLaw = document.fmap(_.identity, m);

    document.toJS(firstLaw).should.eql(document.toJS(document.id()));
  });

  it('holds for the second functor law', function () {
    // fmap (f . g) = fmap f . fmap g
    const inc = v => (_.isInteger(v) ? v + 1 : v);
    const inc2 = v => (_.isInteger(v) ? v + 2 : v);
    const inc3 = _.flow([inc, inc2]);
    const m = document.document({a: 1, b: 2});
    const secondLawLhs = document.fmap(inc3, m);
    const secondLawRhs = document.fmap(inc2, document.fmap(inc, m));

    document.toJS(secondLawLhs).should.eql(document.toJS(secondLawRhs));
  });

  it('holds for the associativity law of monoids', function () {
    // (m1 mappend m2) mappend m3 = m1 mappend (m2 mappend m3)
    const m1 = document.document({a: 1});
    const m2 = document.document({b: 2});
    const m3 = document.document({c: 3});
    const associativityLawLhs = document.mappend(document.mappend(m1, m2), m3);
    const associativityLawRhs = document.mappend(m1, document.mappend(m2, m3));

    document.toJS(associativityLawLhs).should.eql(document.toJS(associativityLawRhs));
  });

  it('holds for the left identity law of monoids', function () {
    // mempty mappend x = x
    const id = document.mempty();
    const m = document.document({a: 1});
    const leftIdentityLaw = document.mappend(id, m);

    document.toJS(leftIdentityLaw).should.eql(docJs({a: 1}));
    Immutable.is(leftIdentityLaw, m);
  });

  it('holds for the right identity law of monoids', function () {
    // x mappend mempty = x
    const id = document.mempty();
    const m = document.document({a: 1});
    const rightIdentityLaw = document.mappend(m, id);

    document.toJS(rightIdentityLaw).should.eql(docJs({a: 1}));
    Immutable.is(rightIdentityLaw, m);
  });
});
